package me.hammerle.supersnuvi.gamelogic;

import java.io.File;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeSet;
import java.util.stream.Collectors;
import me.hammerle.snuviengine.api.Shader;
import me.hammerle.supersnuvi.Game;
import me.hammerle.supersnuvi.entity.Entity;
import me.hammerle.supersnuvi.entity.EntityBuilder;
import me.hammerle.supersnuvi.javafx.Image;
import me.hammerle.supersnuvi.tiles.Location;
import me.hammerle.supersnuvi.tiles.StartTile;
import me.hammerle.supersnuvi.tiles.Tile;
import me.hammerle.supersnuvi.util.CollisionBox;
import me.hammerle.supersnuvi.util.Utils;

public final class Level 
{
    private final boolean worldLoaded;
    private final LevelData data;
    
    private final String fileName;
    private final String name;
    
    private final HashMap<Integer, Entity> entities = new HashMap<>();;
    private final LinkedList<Entity> spawnQueue = new LinkedList<>();
    private Entity hero;
    private int entityCounter = 0;
    
    private boolean shouldReset = false;
    private boolean done = true; // this will be reseted in resetLevel()
    
    private int souls;
    private int maxSouls;
    
    private double time = 0.0;
    private double clockTick = 0.0;
    
    //private int tries = 7;
    
    private TreeSet<Point> spawns = new TreeSet<>();
    
    public Level(File f)
    {
        this.data = new LevelData(f);
        this.worldLoaded = data.load();
        
        if(worldLoaded)
        {
            this.name = data.getString("name", "error");
            
            // debug stuff
            /*if(name.equals("00-Tech Demo"))
            {
                int index = data.getBackgroundIndex() + 1;
                data.setTile(index, 10, 6, 2);
                data.setTile(index, 11, 6, 1);
                
                /*for(int layer = 0; layer <= data.getBackgroundIndex(); layer++)
                {
                    for(int x = 0; x < data.getWidth(); x++)
                    {
                        for(int y = 0; y < data.getHeight(); y++)
                        {
                            int t = data.getTile(layer, x, y);
                            if(t < 16)
                            {
                                data.setTile(layer, x, y, t + 240);
                            }
                            else if(t < 32)
                            {
                                data.setTile(layer, x, y, t + 224);
                            }
                        }
                    }
                }
                
                index--;
                for(int x = 0; x < data.getWidth(); x++)
                {
                    for(int y = 0; y < data.getHeight(); y++)
                    {
                        if(data.getTile(index, x, y) == 128)
                        {
                            data.setTile(index, x, y, -1);
                        }
                    }
                }
                for(int x = 0; x < 5; x++)
                {
                    for(int y = 0; y < 5; y++)
                    {
                        data.setTile(index, x + 6, y + 1, 256 + x + (4 - y) * 5);
                    }
                }
            }*/
            // end debug stuff
            
            maxSouls = 0;
            data.forEachInteractTile((x, y, tile) -> 
            {
                Tile t = Game.get().getTile(tile);
                if(t != null)
                {
                    maxSouls += t.getBottleScore();
                }
                
                if(tile == StartTile.ID)
                {
                    spawns.add(new Point(x, y));
                }
            });
            data.forEachEntity((x, y, tile) -> 
            {            
                if(tile == 1)
                {
                    maxSouls++;
                }
            }, 0, data.getWidth(), 0, data.getHeight());
        }
        else
        {
            this.name = f.getName();
            maxSouls = 0;
        }
        fileName = f.getName();
        
        // there must be at least one spawn
        if(spawns.isEmpty())
        {
            spawns.add(new Point(5, 5));
        }

        resetLevel();
    }
    
    // -------------------------------------------------------------------------
    // basic stuff
    // -------------------------------------------------------------------------

    public String getFileName() 
    {
        return fileName;
    }
    
    public String getName() 
    {
        return name;
    }
    
    public void finishLevel()
    {
        shouldReset = true;
        done = true;
    }
    
    public boolean shouldFinish()
    {
        return done;
    }
    
    public boolean shouldReset()
    {
        return shouldReset;
    }
    
    public void scheduleReset()
    {
        shouldReset = true;
    }
    
    public boolean resetLevel()
    {
        /*boolean dead = false;
        if(!done) // hero just died
        {
            tries--;
            if(tries <= 0)
            {
                tries = 7;
                dead = true;
            }
        }*/
        Game.get().resetTiles();
        data.activateEntities();
        souls = 0;
        time = 0.0;
        shouldReset = false;
        done = false;
        Entity h = spawnHero(true);
        hero = h;
        entities.clear();
        entities.put(entityCounter++, h);
        return false;
    }
    
    public void spawnEntity(Entity ent)
    {
        spawnQueue.add(ent);
    }
    
    public Entity spawnHero(boolean first)
    {
        Entity h = hero;
        Point p;
        if(h == null || hero.getX() < 0 || first)
        {
            // first spawn or out of map, use first spawn
            p = spawns.first();
        }
        else
        {
            // hero is somewhere in the map, getting last spawn
            p = spawns.floor(new Point(Utils.toBlock(hero.getX()), Utils.toBlock(hero.getY())));
            if(p == null)
            {
                p = spawns.first();
            }
        }
        return EntityBuilder.buildHero(this, Utils.toCoord(p.getX()), Utils.toCoord(p.getY()));
    }
    
    public void increaseSouls(int score)
    {
        souls += score;
    }
    
    public int getCurrentBottles()
    {
        return souls;
    }
    
    public int getMaxBottles()
    {
        return maxSouls;
    }
    
    public LevelData getData()
    {
        return data;
    }
    
    public double getTime()
    {
        return time;
    }  

    // -------------------------------------------------------------------------
    // tick
    // -------------------------------------------------------------------------
    
    public void tick()
    {
        if(worldLoaded)
        {
            time += 0.0125;
            
            Game.get().tickTiles();
            
            // entity spawn layer
            /*int startX = renderer.getFirstVisibleBlockX();
            int startY = renderer.getFirstVisibleBlockY();
            int endX = Math.min(renderer.getLastVisibleBlockX() + 1, data.getWidth());
            int endY = Math.min(renderer.getLastVisibleBlockY() + 1, data.getHeight());
            data.forEachEntity((x, y, tile) -> 
            {
                if(tile > 0)
                {
                    data.deactivateEntity(x, y);
 
                    Entity ent = EntityBuilder.fromId(tile, this, Utils.toCoord(x), Utils.toCoord(y));
                    if(ent != null)
                    {
                        entities.put(entityCounter++, ent);
                    }
                }
            }, startX, endX, startY, endY);*/
            
            // doing entity logic first
            entities.values().removeIf(entity -> 
            {
                entity.tick();
                return entity.getHealth().shouldDespawn();
            });
            
            if(!spawnQueue.isEmpty())
            {
                spawnQueue.forEach(ent -> entities.put(entityCounter++, ent));
                spawnQueue.clear();
            }
        }
    }   
    
    public String formatBottles(int bottles)
    {
        char[] c = new char[5];
        if(bottles <= 9)
        {
            c[0] = '0';
            c[1] = (char) (bottles + '0');
        }
        else if(bottles > 99)
        {
            c[0] = 'X';
            c[1] = 'X';
        }
        else
        {
            c[0] = (char) ((bottles / 10) + '0');
            c[1] = (char) ((bottles % 10) + '0');
        }
        c[2] = '/';
        if(maxSouls <= 9)
        {
            c[3] = '0';
            c[4] = (char) (maxSouls + '0');
        }
        else if(maxSouls > 99)
        {
            c[3] = 'X';
            c[4] = 'X';
        }
        else
        {
            c[3] = (char) ((maxSouls / 10) + '0');
            c[4] = (char) ((maxSouls % 10) + '0');
        }
        return new String(c);
    }
    
    public String formatTime(double time)
    {
        if(time == -1.0)
        {
            return "-----";
        }
        else if(time >= 999.9)
        {
            return "999.9";
        }
        return String.format("%05.1f", time);
    }
    
    
    private final static Image AROUND_BACKGROUND = Utils.getImage("gui/around_bars/around_background");
    private final static Image AROUND_FOREGROUND = Utils.getImage("gui/around_bars/around_foreground");
    
    private final static Image ENERGY_EMPTY = Utils.getImage("gui/energy_bar/energy_bar_empty");
    private final static Image ENERGY_FULL = Utils.getImage("gui/energy_bar/energy_bar_full");
    
    private final static Image[] HEALTH_BAR = new Image[]
    {
        Utils.getImage("gui/health_bar/health_bar_0", true),
        Utils.getImage("gui/health_bar/health_bar_1", true),
        Utils.getImage("gui/health_bar/health_bar_2", true),
        Utils.getImage("gui/health_bar/health_bar_3", true),
        Utils.getImage("gui/health_bar/health_bar_4", true),
        Utils.getImage("gui/health_bar/health_bar_5", true),
        Utils.getImage("gui/health_bar/health_bar_6", true),
        Utils.getImage("gui/health_bar/health_bar_7", true)
    };
    
    private final static Image[] LIFE_OVERLAY = new Image[]
    {
        Utils.getImage("gui/health_num/health_num_0", true),
        Utils.getImage("gui/health_num/health_num_1", true),
        Utils.getImage("gui/health_num/health_num_2", true),
        Utils.getImage("gui/health_num/health_num_3", true),
        Utils.getImage("gui/health_num/health_num_4", true),
        Utils.getImage("gui/health_num/health_num_5", true),
        Utils.getImage("gui/health_num/health_num_6", true),
        Utils.getImage("gui/health_num/health_num_7", true)
    };
    
    private final static Image[] TIME_CLOCK = new Image[]
    {
        Utils.getImage("gui/time_icon/time_f1", true),
        Utils.getImage("gui/time_icon/time_f2", true),
        Utils.getImage("gui/time_icon/time_f3", true),
        Utils.getImage("gui/time_icon/time_f4", true),
        Utils.getImage("gui/time_icon/time_f5", true),
        Utils.getImage("gui/time_icon/time_f6", true),
        Utils.getImage("gui/time_icon/time_f7", true),
        Utils.getImage("gui/time_icon/time_f8", true),
        Utils.getImage("gui/time_icon/time_f9", true),
        Utils.getImage("gui/time_icon/time_f10", true),
        Utils.getImage("gui/time_icon/time_f11", true),
        Utils.getImage("gui/time_icon/time_f12", true)
    };
    
    private float getViewX(float x) 
    {
        x -= Shader.getViewWidth() >> 1;
        if(x < 0)
        {
            return 0;
        }
        float max = data.getWidth() * Tile.SIZE - Shader.getViewWidth();
        if(x > max)
        {
            return max;
        }
        return x;
    }
    
    private float getViewY(float y) 
    {
        y -= Shader.getViewHeight() >> 1;
        if(y < 0)
        {
            return 0;
        }
        float max = data.getHeight() * Tile.SIZE - Shader.getViewHeight();
        if(y > max)
        {
            return max;
        }
        return y;
    }
    
    public void render(float lag)
    {
        if(worldLoaded)
        {
            // setting the right view center
            double rWidth = Shader.getViewWidth();
            double rHeight = Shader.getViewHeight();
            
            float centerX = (float) (hero.getX() - hero.getWidth() / 2);
            float centerY = (float) (hero.getY() - hero.getHeight() / 2);
            
            float cameraX = getViewX(centerX);
            float cameraY = getViewY(centerY);
            
            Shader.translateTo(-cameraX, -cameraY);
            Shader.updateMatrix();
            
            Shader.setColorEnabled(true);
            Shader.setTextureEnabled(false);
            Shader.getColorRenderer().drawRectangle(centerX - 5, centerY - 5, centerX + 5, centerY + 5, 0xFFFF0000);
            
            /*data.forEachInteractTile((x, y, tile) -> 
            {
                if(tile <= 0)
                {
                    return;
                }
                x *= Tile.SIZE;
                y *= Tile.SIZE;
                Shader.getColorRenderer().drawRectangle(x, y, x + Tile.SIZE, y  + Tile.SIZE, 0xFFFFFF00);
            });*/
            
            /*renderer.setViewCenter(centerX, centerY);
            
            int startX = renderer.getFirstVisibleBlockX();
            int startY = renderer.getFirstVisibleBlockY();
            int endX = Math.min(renderer.getLastVisibleBlockX() + 1, data.getWidth());
            int endY = Math.min(renderer.getLastVisibleBlockY() + 1, data.getHeight());*/
            
            /*data.forEachBackground((x, y, tile) ->
            {
                if(tile != -1)
                {
                    Tile t = state.getTile(tile);
                    t.preRender(renderer, x, y);
                    renderer.drawBlockImage(t.getImage(x, y), x, y, t.getRenderOffsetX(), t.getRenderOffsetY());
                    t.postRender(renderer, x, y);
                }
            }, startX, endX, startY, endY);
            
            entities.values().forEach(en -> en.getAnimator().render(renderer));
            
            data.forEachForeground((x, y, tile) ->
            {
                if(tile != -1)
                {
                    Tile t = state.getTile(tile);
                    t.preRender(renderer, x, y);
                    renderer.drawBlockImage(t.getImage(x, y), x, y, t.getRenderOffsetX(), t.getRenderOffsetY());
                    t.postRender(renderer, x, y);
                }
            }, startX, endX, startY, endY);
            
            // menu rendering
            // |-------------------------------------| 
            // | B00/00 | ENERGYENERGY | HPHP | TIME |
            // |-------------------------------------| 
            renderer.prepareTextDrawing(255, 255, 255, 1.0, 40);
            {
                double scale = renderer.getHeight() / 300;
                
                double w = AROUND_BACKGROUND.getWidth() * scale;
                double h = AROUND_BACKGROUND.getHeight() * scale;
                
                double x = renderer.getWidth() - w;
                double y = 0;
                
                // gui background
                renderer.drawFixedImage(AROUND_BACKGROUND, x, y, w, h);
                
                // health mirror    
                int hFrame = (int) (hero.getHealth().getHealthPercent() * (HEALTH_BAR.length - 1));
                renderer.drawFixedImage(HEALTH_BAR[hFrame], x + 78 * scale, y + 8 * scale, 
                        HEALTH_BAR[hFrame].getWidth() * scale, HEALTH_BAR[hFrame].getHeight() * scale);
                
                // energy
                renderer.drawFixedImage(ENERGY_EMPTY, x + 8 * scale, y + 12 * scale, 
                        ENERGY_EMPTY.getWidth() * scale, ENERGY_EMPTY.getHeight() * scale);
                
                double wEnergy = ENERGY_FULL.getWidth() * hero.getEnergy().getEnergyPercent();
                double wEnergyOffset = ENERGY_FULL.getWidth() - wEnergy;
                renderer.drawFixedImagePart(ENERGY_FULL, 
                        wEnergyOffset, 0, wEnergy, ENERGY_FULL.getHeight(),
                        x + (wEnergyOffset + 8) * scale, y + 12 * scale,
                        wEnergy * scale, ENERGY_FULL.getHeight() * scale);
                
                // gui foreground
                renderer.drawFixedImage(AROUND_FOREGROUND, x, y, w, h);   
                
                // tries / health go over foreground
                renderer.drawFixedImage(LIFE_OVERLAY[hFrame], x + 89 * scale, y + 52 * scale,
                        LIFE_OVERLAY[hFrame].getWidth() * scale, LIFE_OVERLAY[hFrame].getHeight() * scale);
                
                
                // grey background
                double boxHeight = renderer.getTextHeight(2) + 10;
                double boxWidth = renderer.getTextWidth(6) + 10;
                
                renderer.save();
                {
                    renderer.setFillColor(0, 0, 0, 0.5);
                    renderer.fillRectangle(0, 0, boxWidth, boxHeight);  
                }
                renderer.restore();

                // soul / time rendering
                x = 5;
                y = 5;
                w = renderer.getTextWidth(1);
                h = renderer.getTextHeight(1);
                
                // soul rendering
                renderer.drawFixedImagePart(BottledSoulTile.IMAGE[0], 14, 8, 34, 52, x, y, w, h);
                renderer.drawText(x + renderer.getTextWidth(1), y, formatBottles(souls));

                y += renderer.getTextHeight(1);
                
                // time rendering
                clockTick += 0.05;
                if(clockTick >= TIME_CLOCK.length)
                {
                    clockTick = 0.0;
                }
                renderer.drawFixedImagePart(TIME_CLOCK[(int) clockTick], 0, 0, 32, 32, x, y, w, h);
                renderer.drawText(x + renderer.getTextWidth(1), y, formatTime(time));
            }
            renderer.stopTextDrawing();*/
        }
    }
    
    // -------------------------------------------------------------------------
    // collision box, interaction layer
    // -------------------------------------------------------------------------
    
    public Tile getInteractionTile(int x, int y)
    {
        int i = data.getInteractionTile(x, y);
        if(i == -1)
        {
            return Game.FALLBACK_TILE;
        }
        return Game.get().getTile(i);
    }
    
    public CollisionBox getTileMovementBox(int x, int y)
    {
        int i = data.getInteractionTile(x, y);
        if(i == -1)
        {
            return CollisionBox.NULL_BOX;
        }
        Tile tile = Game.get().getTile(i);
        return tile.getMovementBox(x, y).reset().offset(Utils.toCoord(x), Utils.toCoord(y));
    }
    
    public LinkedList<Location> getTilesInMovementOf(CollisionBox cb)
    {
        LinkedList<Location> boxes = new LinkedList<>();
        int startX = Utils.toBlock(cb.getMinX());
        int endX = Utils.toBlock(cb.getMaxX());
        int startY = Utils.toBlock(cb.getMinY());
        int endY = Utils.toBlock(cb.getMaxY());
        
        for(int x = startX; x <= endX; x++)
        {
            for(int y = startY; y <= endY; y++)
            {
                if(getTileMovementBox(x, y).intersects(cb))
                {
                    boxes.add(new Location(getInteractionTile(x, y), x, y));
                }
            }
        }
        return boxes;
    }
    
    public CollisionBox getTileCollisionBox(int x, int y)
    {
        int i = data.getInteractionTile(x, y);
        if(i == -1)
        {
            return CollisionBox.NULL_BOX;
        }
        Tile tile = Game.get().getTile(i);
        return tile.getCollisionBox(x, y).reset().offset(Utils.toCoord(x), Utils.toCoord(y));
    }
    
    public LinkedList<Location> getTilesCollidingWith(CollisionBox cb)
    {
        LinkedList<Location> boxes = new LinkedList<>();
        int startX = Utils.toBlock(cb.getMinX());
        int endX = Utils.toBlock(cb.getMaxX());
        int startY = Utils.toBlock(cb.getMinY());
        int endY = Utils.toBlock(cb.getMaxY());
        
        for(int x = startX; x <= endX; x++)
        {
            for(int y = startY; y <= endY; y++)
            {
                if(getTileCollisionBox(x, y).intersects(cb))
                {
                    boxes.add(new Location(getInteractionTile(x, y), x, y));
                }
            }
        }
        return boxes;
    }
    
    public List<Entity> getEntitiesCollidingWith(Entity not, CollisionBox cb)
    {
        return entities.values().stream().filter(ent -> ent != not && ent.getBox().intersects(cb)).collect(Collectors.toList());
    }
    
    public LinkedList<CollisionBox> getMovementBoxesAt(Entity not, CollisionBox cb)
    {
        LinkedList<CollisionBox> boxes = new LinkedList<>(entities.values().stream()
                            .filter(ent -> ent != not && ent.getMovement().useCollisionBox() && ent.getBox().intersects(cb))
                            .map(ent -> ent.getBox())
                            .collect(Collectors.toList()));
        int startX = Utils.toBlock(cb.getMinX());
        int endX = Utils.toBlock(cb.getMaxX());
        int startY = Utils.toBlock(cb.getMinY());
        int endY = Utils.toBlock(cb.getMaxY());
        
        CollisionBox box;
        for(int x = startX; x <= endX; x++)
        {
            for(int y = startY; y <= endY; y++)
            {
                box = getTileMovementBox(x, y);
                if(box.intersects(cb))
                {
                    boxes.add(box.copy());
                }
            }
        }
        return boxes;
    }
    
    public Entity getHero()
    {
        return hero;
    }
}
